/*
** Module   : DynamicQuery.java
** Abstract : Base class for queries which fetch records progressively
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20051215   @23741 Created initial version. Primary purpose at
**                           this time is to centralize logic and data
**                           related to scrollable cursor support.
** 002 ECF 20060104   @23823 Changed hierarchy. Extends AbstractQuery base
**                           class.
** 003 ECF 20060315   @25076 Added safety code to setScrolling. It is now
**                           tolerant of multiple invocations; a query
**                           cursor is created only on the first call.
** 004 ECF 20060501   @25914 Added isPreselect method. Indicates query's
**                           preselection behavior.
** 005 ECF 20060801   @28350 Added abstract getBreakValue() method.
** 006 ECF 20061102   @30957 Added support for reposition notification and
**                           auto-fetch features.
** 007 ECF 20061206   @31586 Integrated changes to AbstractQuery's
**                           afterReposition() signature. Leverage new
**                           return value from Cursor's reposition()
**                           variant which accepts an array of IDs, to
**                           ensure reposition listeners are notified of
**                           default reposition in the event of an error.
** 008 ECF 20070518   @33680 Implement RecordChangeListener interface.
**                           This object is now registered as a listener
**                           with the ChangeBroker.
** 009 ECF 20070522   @33723 Fixed stateChanged() method. Removed throw of
**                           IllegalStateException. The logic for this
**                           exception was incorrect.
** 010 ECF 20080122   @36886 Added resetScrolling(). Replaces cursor with
**                           a new instance.
** 011 ECF 20080201   @36999 Fixed currentRow(). Return unknown if query
**                           is not scrolling, rather than throwing an
**                           exception.
** 012 SVL 20080604   @38532 Modified according to the new P2JQuery interface.
** 013 ECF 20080729   @39319 Added rowid support. Refactored reposition*()
**                           methods.
** 014 ECF 20081106   @40322 Fixed stateChanged(), added _isOffEnd().
** 015 ECF 20090521   @42979 Changes to fix incorrect behavior. Removed
**                           RecordChangeListener interface and stateChanged()
**                           methods. Track how many times the current result
**                           row has changed and report this as the current
**                           row number for non-scrolling queries.
** 016 ECF 20090815   @43632 Explicitly invoke superclass default c'tor in
**                           this default c'tor.
** 017 SVL 20140909          Added isNativelyPreselect and isScrolling.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * Abstract base class for query classes which dynamically navigate and
 * retrieve results, and which must support results list scrolling.
 * <p>
 * The primary purpose of this class is to manage a query cursor for results
 * list scrolling and repositioning, and to provide an API to client code to
 * perform these tasks, and to interrogate the cursor for state data.
 * <p>
 * This class implements the <code>RecordChangeListener</code> interface.
 * When the {@link #setScrolling} method is invoked, this object is registered
 * as a listener with the context-local {@link ChangeBroker} instance, so
 * that it can receive notifications of record deletions.  If the deleted
 * record was in this query's cache, the cache is expired.  Subclasses must
 * implement the {@link
 * com.goldencode.p2j.persist.event.RecordChangeListener#recordBuffers
 * recordBuffers()} method.
 */
abstract class DynamicQuery
extends AbstractQuery
implements Scrollable
{
   /** Number of times the query's position has moved (non-scrolling only) */
   private int moveCount = 0;
   
   /** Cursor which scrolls query results */
   protected Cursor cursor = null;
   
   /**
    * Default constructor.
    */
   protected DynamicQuery()
   {
      super();
   }
   
   /**
    * Create a query cursor to enable results list scrolling/repositioning.
    * Only one cursor is created, even if this method is invoked multiple
    * times.
    */
   public void setScrolling()
   {
      if (!isScrolling())
      {
         resetScrolling();
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   public void repositionByID(recid id)
   {
      verifyScrolling();
      ErrorConditionException error = null;
      boolean permitFetch = true;
      
      try
      {
         permitFetch = cursor.repositionByID(id);
      }
      catch (ErrorConditionException exc)
      {
         permitFetch = false;
         error = exc;
      }
      finally
      {
         afterReposition(permitFetch);
      }
      
      if (error != null)
      {
         throw error;
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    * 
    * @throws  ErrorConditionException
    *          if any of the specified IDs represents the unknown value.
    */
   public void repositionByID(rowid id1, rowid...joinIDs)
   {
      verifyScrolling();
      ErrorConditionException error = null;
      boolean permitFetch = true;
      
      try
      {
         permitFetch = cursor.repositionByID(id1, joinIDs);
      }
      catch (ErrorConditionException exc)
      {
         permitFetch = false;
         error = exc;
      }
      finally
      {
         afterReposition(permitFetch);
      }
      
      if (error != null)
      {
         throw error;
      }
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>row</code> represents unknown value.
    */
   public void reposition(NumberType row)
   {
      verifyScrolling();
      cursor.reposition(row);
      afterReposition(true);
   }
   
   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void reposition(int row)
   {
      verifyScrolling();
      cursor.reposition(row);
      afterReposition(true);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void forward(NumberType rows)
   {
      verifyScrolling();
      cursor.forward(rows);
      afterReposition(true);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void forward(int rows)
   {
      verifyScrolling();
      cursor.forward(rows);
      afterReposition(true);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void backward(NumberType rows)
   {
      verifyScrolling();
      cursor.backward(rows);
      afterReposition(true);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void backward(int rows)
   {
      verifyScrolling();
      cursor.backward(rows);
      afterReposition(true);
   }
   
   /**
    * Get the <b><i>current</i></b> size of the cached results list.  However,
    * since this cache is built progressively as the query is scrolled,
    * repositioned, and naturally navigated, this <b>does not</b> necessarily
    * represent the full count of results which may satisfy the query.
    *
    * @return  Current results list size.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer size()
   {
      verifyScrolling();
      
      return cursor.size();
   }
   
   /**
    * Get the 1-based index of the current row in the result set.  This is
    * the index of the entry at or before which the cursor currently is
    * positioned.
    *
    * @return  Current row or the unknown value if the current row cannot be
    *          determined.  The current row cannot be determined if the
    *          results list is empty, or if the cursor currently is in
    *          backward scroll mode.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer currentRow()
   {
      return currentRowImpl();
   }
   
   /**
    * Get the 1-based index of the current row in the result set.  This is
    * the index of the entry at or before which the cursor currently is
    * positioned.
    *
    * @return  Current row or the unknown value if the current row cannot be
    *          determined.  The current row cannot be determined if the
    *          results list is empty.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer currentRowImpl()
   {
      if (!isScrolling())
      {
         return (moveCount > 0 ? new integer(moveCount) : new integer());
      }
      
      return cursor.currentRow();
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public boolean _isOffEnd()
   {
      if (isScrolling())
      {
         return cursor.isOffEnd(null);
      }
      
      return !OffEnd.NONE.equals(getOffEnd());
   }
   
   /**
    * Indicate whether this query preselects all of its results.
    *
    * @return  <code>false</code>.
    */
   public boolean isPreselect()
   {
      return false;
   }

   /**
    * Indicate whether this query is a preselect query by its nature.
    *
    * @return  <code>true</code>.
    */
   public boolean isNativelyPreselect()
   {
      return false;
   }
   
   /**
    * Create a query cursor to enable results list scrolling/repositioning.
    */
   protected void resetScrolling()
   {
      cursor = new Cursor(this);
   }
   
   /**
    * Verify that this query is a scrolling query;  that is, that it has an
    * associated cursor.
    *
    * @throws  UnsupportedOperationException
    *          if a cursor has not been associated with this query.
    */
   protected void verifyScrolling()
   {
      if (!isScrolling())
      {
         throw new UnsupportedOperationException(
            "Cannot access cursor for a non-scrolling query");
      }
   }

   /**
    * Indicate whether this query's <code>scrolling</code> flag has been set.
    *
    * @return  <code>true</code> if this is a scrolling query;  else
    *          <code>false</code>.
    */
   protected boolean isScrolling()
   {
      return cursor != null;
   }
   
   /**
    * Increment the counter of how many times the current result row of this
    * query has changed.
    */
   protected void incrementMoves()
   {
      moveCount++;
   }
   
   /**
    * Return an object which represents a property value which defines the
    * end of a <i>sort band</i>.  A <i>sort band</i> is a series of one or
    * more records which share the same value for a particular property,
    * where that property is used as the coarsest sorting criterion defined
    * by this query's <code>order by</code> clause.
    * <p>
    * If the break value for this query currently is non-<code>null</code>,
    * this indicates that the most recently retrieved record has crossed a
    * <i>sort band</i> boundary, and that a new band has begun.  This value
    * will only be non-<code>null</code> at the point at which such a record
    * is found, and will be <code>null</code> for all other records.
    * <p>
    * Note:  break values are only tracked for invocations of
    * <code>next</code> and <code>previous</code>.  Other retrievals will
    * results in break value being <code>null</code>.
    *
    * @return  Break value if dynamic query crossed a sort band boundary,
    *          else <code>null</code>.
    */
   protected abstract Object getBreakValue();
}
